#include "oulib.h"
#include "stdlib.h"

extern int uart_send_general_buffered(char c, FILE *fp);
extern int uart_recv_general_buffered(FILE *fp);


// uint8_t serial_init_buffered(uint8_t port, unsigned long baud_rate,
// 			   uint8_t buf_size_in, uint8_t buf_size_out)  
//
//  Experimental serial initialization that supports both single- and
// multi-port atmega chips.
//  If multi-port, then we create the necessary structures to
// keep track of which port is which.  Ultimately, this
// structure will be used to do buffering as well.
//
//  Return = 0 if no errors
//         = -1 if an incorrect port was specified

/**
   Initialize RS232 buffered serial support for standard I/O.  This will allow you to use
   the standard I/O facilities (getchar(), printf(), scanf(), etc) - see the libc documentation
   for more information.

   <P>
   Assumptions: the calling function will enable global interrupts, and each port is only
   initialized once.

   <P>

   Note: characters received that do not fit in the input buffer will be dropped on the floor.
   Outgoing characters that do not fit into the buffer will wait until there is room
   in the buffer.

   <P>
   @param port The serial port to initialize.  Port 0 is supported on all chips.  Port 1 is
   supported only on multi-port chips.
   @param baud_rate The transceiver clock rate.
   @param buf_size_in The number of bytes allocated to the input buffer.  Must be 2 or more.
   @param buf_size_out The number of bytes allocate to the output buffer.  Must not be 1
   (0 indicates that no buffer is used).

   @return A pointer to the newly created FILE structure (a valid file pointer).
   This file pointer will only be needed for certain operations (functions such as printf()
   and getchar() default to the port 0 file pointer). <BR>
   NULL is returned if an error occurred.
   

   @author  Andrew H. Fagg

   \ingroup serial
 */

FILE* serial_init_buffered(uint8_t port, unsigned long baud_rate,
			   uint8_t buf_size_in, uint8_t buf_size_out)  
{
  //#ifndef OULIB_EXP
  ////buf_size_out = 0;  // Right now, we do not buffer the output side
  //  if(buf_size_in + buf_size_out <= 1) {
  //    return(NULL);
  //  };
  //#else
  // Need at least 2 elements in the buffer (this will allow us to
  //  have up to one element).
  // Output buffer may have 0 elements
  
  if(buf_size_in < 2 || buf_size_out == 1) {
    return(NULL);
  };
  //#endif
  // This structure will be linked to the FILE structure
  struct OU_FILE_BUF *ou_fp = (struct OU_FILE_BUF *)
    malloc(sizeof(struct OU_FILE_BUF));
  ou_fp->port = port;
  ou_fp->buf_size_in = buf_size_in;
  ou_fp->buf_size_out = buf_size_out;
  ou_fp->buf_in_front = 
    ou_fp->buf_in_back = 0;
  ou_fp->buf_out_front = 
    ou_fp->buf_out_back = 0;

  // To save space, we will have one set of characters for
  //  the two buffers
  ou_fp->buffer = (char *) malloc(buf_size_in+buf_size_out);
    
  FILE *fp;

  //  Compute baud rate divisor.  We assume that the user
  // provides a reasonable request
  unsigned int baud = (F_CPU / (16 * baud_rate)) - 1;

  //  Depending on the specified port, we will configure
  // different registers
  switch(port)
    {
    case 0:
      UCSRB = (1<<RXEN)|(1<<TXEN);
      UBRRH = (unsigned char) (baud>>8);
      UBRRL = (unsigned char) baud;
      fp = fdevopen(uart_send_general_buffered, uart_recv_general_buffered);
      //  Link the structure to the FILE structure
      fdev_set_udata(fp, ou_fp);
      // Remember this stream
      ou_file_buf[0] = ou_fp;
      // Enable the receive interrupt
      serial0_receive_enable();
      return(fp);
      break;
#if defined(SERIAL_NPORTS_2) | defined(SERIAL_NPORTS_4)
    case 1:
      // Devices with at least 2 ports
      UCSR1B = (1<<RXEN)|(1<<TXEN);
      UBRR1H = (unsigned char) (baud>>8);
      UBRR1L = (unsigned char) baud;
      fp = fdevopen(uart_send_general_buffered, uart_recv_general_buffered);
      fdev_set_udata(fp, ou_fp);
      // Remember this stream
      ou_file_buf[1] = ou_fp;
      serial1_receive_enable();
      return(fp);
      break;
#endif
#if defined(SERIAL_NPORTS_4)
      // Devices with at least 4 ports
    case 2:
      // Port 2
      UCSR2B = (1<<RXEN)|(1<<TXEN);
      UBRR2H = (unsigned char) (baud>>8);
      UBRR2L = (unsigned char) baud;
      fp = fdevopen(uart_send_general_buffered, uart_recv_general_buffered);
      fdev_set_udata(fp, ou_fp);
      // Remember this stream
      ou_file_buf[2] = ou_fp;
      serial2_receive_enable();
      return(fp);
      break;
    case 3:
      // Port 3
      UCSR3B = (1<<RXEN)|(1<<TXEN);
      UBRR3H = (unsigned char) (baud>>8);
      UBRR3L = (unsigned char) baud;
      fp = fdevopen(uart_send_general_buffered, uart_recv_general_buffered);
      fdev_set_udata(fp, ou_fp);
      // Remember this stream
      ou_file_buf[3] = ou_fp;
      serial3_receive_enable();
      return(fp);
      break;
#endif
    default:
      // Indicate an error
      if(ou_fp->buffer)
	free(ou_fp->buffer);
      free(ou_fp);
      return(NULL);
    }
}

